from django import forms

from .. import FraudStatus
from .. import PaymentStatus
from ..forms import PaymentForm


class DummyForm(PaymentForm):
    RESPONSE_CHOICES = (
        ("3ds-disabled", "3DS disabled"),
        ("3ds-redirect", "3DS redirect"),
        ("failure", "Gateway connection error"),
        ("payment-error", "Gateway returned unsupported response"),
    )
    status = forms.ChoiceField(choices=PaymentStatus.CHOICES)
    fraud_status = forms.ChoiceField(choices=FraudStatus.CHOICES)
    gateway_response = forms.ChoiceField(choices=RESPONSE_CHOICES)
    verification_result = forms.ChoiceField(
        choices=PaymentStatus.CHOICES, required=False
    )

    def clean(self):
        cleaned_data = super().clean()
        gateway_response = cleaned_data.get("gateway_response")
        verification_result = cleaned_data.get("verification_result")
        if gateway_response == "3ds-redirect" and not verification_result:
            raise forms.ValidationError(
                "When 3DS is enabled you must set post validation status"
            )
        return cleaned_data



from decimal import Decimal
from typing import NamedTuple
from typing import Optional

from django.apps import apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import pgettext_lazy

from payments import version  # type: ignore

__version__ = version.version


class PurchasedItem(NamedTuple):
    """A single item in a purchase."""

    name: str
    quantity: int
    price: Decimal
    currency: str
    sku: str
    tax_rate: Optional[Decimal] = None


class PaymentStatus:
    WAITING = "waiting"
    PREAUTH = "preauth"
    CONFIRMED = "confirmed"
    REJECTED = "rejected"
    REFUNDED = "refunded"
    ERROR = "error"
    INPUT = "input"

    CHOICES = [
        (WAITING, pgettext_lazy("payment status", "Waiting for confirmation")),
        (PREAUTH, pgettext_lazy("payment status", "Pre-authorized")),
        (CONFIRMED, pgettext_lazy("payment status", "Confirmed")),
        (REJECTED, pgettext_lazy("payment status", "Rejected")),
        (REFUNDED, pgettext_lazy("payment status", "Refunded")),
        (ERROR, pgettext_lazy("payment status", "Error")),
        (INPUT, pgettext_lazy("payment status", "Input")),
    ]


class FraudStatus:
    UNKNOWN = "unknown"
    ACCEPT = "accept"
    REJECT = "reject"
    REVIEW = "review"

    CHOICES = [
        (UNKNOWN, pgettext_lazy("fraud status", "Unknown")),
        (ACCEPT, pgettext_lazy("fraud status", "Passed")),
        (REJECT, pgettext_lazy("fraud status", "Rejected")),
        (REVIEW, pgettext_lazy("fraud status", "Review")),
    ]


class RedirectNeeded(Exception):
    pass


class PaymentError(Exception):
    def __init__(self, message, code=None, gateway_message=None):
        super().__init__(message)
        self.code = code
        self.gateway_message = gateway_message


class ExternalPostNeeded(Exception):
    pass


def get_payment_model():
    """
    Return the Payment model that is active in this project
    """
    try:
        app_label, model_name = settings.PAYMENT_MODEL.split(".")
    except (ValueError, AttributeError):
        raise ImproperlyConfigured(
            "PAYMENT_MODEL must be of the form " '"app_label.model_name"'
        )
    payment_model = apps.get_model(app_label, model_name)
    if payment_model is None:
        msg = (
            'PAYMENT_MODEL refers to model "%s" that has not been installed'
            % settings.PAYMENT_MODEL
        )
        raise ImproperlyConfigured(msg)
    return payment_model



import hashlib

from django import forms

from .. import PaymentStatus

NEW = "new"
PROCESSING = "processing"
COMPLETED = "completed"
REJECTED = "rejected"
PROCESSING_REALIZATION_WAITING = "processing_realization_waiting"
PROCESSING_REALIZATION = "processing_realization"


class ProcessPaymentForm(forms.Form):
    id = forms.CharField(required=False)
    operation_number = forms.CharField(required=False)
    operation_type = forms.CharField(required=False)
    operation_status = forms.CharField(required=False)
    operation_amount = forms.CharField(required=False)
    operation_currency = forms.CharField(required=False)
    operation_withdrawal_amount = forms.CharField(required=False)
    operation_commission_amount = forms.CharField(required=False)
    is_completed = forms.CharField(required=False)
    operation_original_amount = forms.CharField(required=False)
    operation_original_currency = forms.CharField(required=False)
    operation_datetime = forms.CharField(required=False)
    operation_related_number = forms.CharField(required=False)
    control = forms.CharField(required=False)
    description = forms.CharField(required=False)
    email = forms.CharField(required=False)
    p_info = forms.CharField(required=False)
    p_email = forms.CharField(required=False)
    credit_card_issuer_identification_number = forms.CharField(required=False)
    credit_card_masked_number = forms.CharField(required=False)
    credit_card_brand_codename = forms.CharField(required=False)
    credit_card_brand_code = forms.CharField(required=False)
    credit_card_id = forms.CharField(required=False)
    channel = forms.CharField(required=False)
    channel_country = forms.CharField(required=False)
    geoip_country = forms.CharField(required=False)
    signature = forms.CharField(required=True)

    def __init__(self, payment, pin, **kwargs):
        super().__init__(**kwargs)
        self.pin = pin
        self.payment = payment

    def clean(self):
        cleaned_data = super().clean()
        if not self.errors:
            key_vars = (
                self.pin,
                cleaned_data.get("id", ""),
                cleaned_data.get("operation_number", ""),
                cleaned_data.get("operation_type", ""),
                cleaned_data.get("operation_status", ""),
                cleaned_data.get("operation_amount", ""),
                cleaned_data.get("operation_currency", ""),
                cleaned_data.get("operation_withdrawal_amount", ""),
                cleaned_data.get("operation_commission_amount", ""),
                cleaned_data.get("is_completed", ""),
                cleaned_data.get("operation_original_amount", ""),
                cleaned_data.get("operation_original_currency", ""),
                cleaned_data.get("operation_datetime", ""),
                cleaned_data.get("operation_related_number", ""),
                cleaned_data.get("control", ""),
                cleaned_data.get("description", ""),
                cleaned_data.get("email", ""),
                cleaned_data.get("p_info", ""),
                cleaned_data.get("p_email", ""),
                cleaned_data.get("credit_card_issuer_identification_number", ""),
                cleaned_data.get("credit_card_masked_number", ""),
                cleaned_data.get("credit_card_brand_codename", ""),
                cleaned_data.get("credit_card_brand_code", ""),
                cleaned_data.get("credit_card_id", ""),
                cleaned_data.get("channel", ""),
                cleaned_data.get("channel_country", ""),
                cleaned_data.get("geoip_country", ""),
            )
            key = "".join(key_vars)
            sha256 = hashlib.sha256()
            sha256.update(key.encode("utf-8"))
            key_hash = sha256.hexdigest()
            if key_hash != self.cleaned_data["signature"]:
                self._errors["signature"] = self.error_class(["Bad hash"])
            if int(cleaned_data["control"]) != self.payment.id:
                self._errors["control"] = self.error_class(["Bad payment id"])
        return cleaned_data

    def save(self, *args, **kwargs):
        status = self.cleaned_data["operation_status"]
        self.payment.transaction_id = self.cleaned_data["operation_number"]
        self.payment.save()
        if status == COMPLETED:
            self.payment.captured_amount = self.payment.total
            self.payment.change_status(PaymentStatus.CONFIRMED)
        elif status == REJECTED:
            self.payment.change_status(PaymentStatus.REJECTED)



from unittest import TestCase
from unittest.mock import MagicMock
from urllib.error import URLError
from urllib.parse import urlencode

from .. import FraudStatus
from .. import PaymentError
from .. import PaymentStatus
from .. import RedirectNeeded
from . import DummyProvider

VARIANT = "dummy-3ds"


class Payment:
    id = 1
    variant = VARIANT
    currency = "USD"
    total = 100
    status = PaymentStatus.WAITING
    fraud_status = ""

    def get_process_url(self):
        return "http://example.com"

    def get_failure_url(self):
        return "http://cancel.com"

    def get_success_url(self):
        return "http://success.com"

    def change_status(self, new_status):
        self.status = new_status

    def change_fraud_status(self, fraud_status):
        self.fraud_status = fraud_status


class TestDummy3DSProvider(TestCase):
    def setUp(self):
        self.payment = Payment()

    def test_process_data_supports_verification_result(self):
        provider = DummyProvider()
        verification_status = PaymentStatus.CONFIRMED
        request = MagicMock()
        request.GET = {"verification_result": verification_status}
        response = provider.process_data(self.payment, request)
        self.assertEqual(self.payment.status, verification_status)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response["location"], self.payment.get_success_url())

    def test_process_data_redirects_to_success_on_payment_success(self):
        self.payment.status = PaymentStatus.PREAUTH
        provider = DummyProvider()
        request = MagicMock()
        request.GET = {}
        response = provider.process_data(self.payment, request)
        self.assertEqual(response["location"], self.payment.get_success_url())

    def test_process_data_redirects_to_failure_on_payment_failure(self):
        self.payment.status = PaymentStatus.REJECTED
        provider = DummyProvider()
        request = MagicMock()
        request.GET = {}
        response = provider.process_data(self.payment, request)
        self.assertEqual(response["location"], self.payment.get_failure_url())

    def test_provider_supports_non_3ds_transactions(self):
        provider = DummyProvider()
        data = {
            "status": PaymentStatus.PREAUTH,
            "fraud_status": FraudStatus.UNKNOWN,
            "gateway_response": "3ds-disabled",
            "verification_result": "",
        }
        with self.assertRaises(RedirectNeeded) as exc:
            provider.get_form(self.payment, data)
            self.assertEqual(exc.args[0], self.payment.get_success_url())

    def test_provider_raises_verification_result_needed_on_success(self):
        provider = DummyProvider()
        data = {
            "status": PaymentStatus.WAITING,
            "fraud_status": FraudStatus.UNKNOWN,
            "gateway_response": "3ds-redirect",
        }

        form = provider.get_form(self.payment, data)
        self.assertFalse(form.is_valid())

    def test_provider_supports_3ds_redirect(self):
        provider = DummyProvider()
        verification_result = PaymentStatus.CONFIRMED
        data = {
            "status": PaymentStatus.WAITING,
            "fraud_status": FraudStatus.UNKNOWN,
            "gateway_response": "3ds-redirect",
            "verification_result": verification_result,
        }
        params = urlencode({"verification_result": verification_result})
        expected_redirect = f"{self.payment.get_process_url()}?{params}"

        with self.assertRaises(RedirectNeeded) as exc:
            provider.get_form(self.payment, data)
            self.assertEqual(exc.args[0], expected_redirect)

    def test_provider_supports_gateway_failure(self):
        provider = DummyProvider()
        data = {
            "status": PaymentStatus.WAITING,
            "fraud_status": FraudStatus.UNKNOWN,
            "gateway_response": "failure",
            "verification_result": "",
        }
        with self.assertRaises(URLError):
            provider.get_form(self.payment, data)

    def test_provider_raises_redirect_needed_on_success(self):
        provider = DummyProvider()
        data = {
            "status": PaymentStatus.PREAUTH,
            "fraud_status": FraudStatus.UNKNOWN,
            "gateway_response": "3ds-disabled",
            "verification_result": "",
        }
        with self.assertRaises(RedirectNeeded) as exc:
            provider.get_form(self.payment, data)
            self.assertEqual(exc.args[0], self.payment.get_success_url())

    def test_provider_raises_redirect_needed_on_failure(self):
        provider = DummyProvider()
        data = {
            "status": PaymentStatus.ERROR,
            "fraud_status": FraudStatus.UNKNOWN,
            "gateway_response": "3ds-disabled",
            "verification_result": "",
        }
        with self.assertRaises(RedirectNeeded) as exc:
            provider.get_form(self.payment, data)
            self.assertEqual(exc.args[0], self.payment.get_failure_url())

    def test_provider_raises_payment_error(self):
        provider = DummyProvider()
        data = {
            "status": PaymentStatus.PREAUTH,
            "fraud_status": FraudStatus.UNKNOWN,
            "gateway_response": "payment-error",
            "verification_result": "",
        }
        with self.assertRaises(PaymentError):
            provider.get_form(self.payment, data)

    def test_provider_switches_payment_status_on_get_form(self):
        provider = DummyProvider()
        provider.get_form(self.payment, data={})
        self.assertEqual(self.payment.status, PaymentStatus.INPUT)



from django.forms.utils import flatatt
from django.forms.widgets import HiddenInput
from django.forms.widgets import Input
from django.utils.encoding import force_str
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _


class StripeCheckoutWidget(Input):
    is_hidden = True

    def __init__(self, provider, payment, *args, **kwargs):
        attrs = kwargs.get("attrs", {})
        kwargs["attrs"] = {
            "class": "stripe-button",
            "data-key": provider.public_key,
            "data-image": provider.image,
            "data-name": provider.name,
            "data-email": payment.billing_email,
            "data-description": payment.description or _("Total payment"),
            # Stripe accepts cents
            "data-amount": int(payment.total * 100),
            "data-currency": payment.currency,
        }
        kwargs["attrs"].update(attrs)
        super().__init__(*args, **kwargs)

    def render(self, name, value, attrs=None, renderer=None):
        if value is None:
            value = ""
        final_attrs = dict(attrs or {}, src="https://checkout.stripe.com/checkout.js")
        final_attrs.update(self.attrs)
        del final_attrs["id"]
        if value != "":
            # Only add the 'value' attribute if a value is non-empty.
            final_attrs["value"] = force_str(self.format_value(value))
        return format_html("<script{0}></script>", flatatt(final_attrs))


class StripeWidget(HiddenInput):
    class Media:
        js = ["https://js.stripe.com/v2/", "js/payments/stripe.js"]

    def __init__(self, attrs=None):
        attrs = dict(attrs or {}, id="id_stripe_token")
        super().__init__(attrs)



import json
from typing import Iterable
from typing import Union
from uuid import uuid4

from django.db import models
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from phonenumber_field.modelfields import PhoneNumberField

from . import FraudStatus
from . import PaymentStatus
from . import PurchasedItem
from .core import provider_factory


class PaymentAttributeProxy:
    def __init__(self, payment):
        self._payment = payment
        super().__init__()

    def __getattr__(self, item):
        data = json.loads(self._payment.extra_data or "{}")
        try:
            return data[item]
        except KeyError as e:
            raise AttributeError(*e.args)

    def __setattr__(self, key, value):
        if key == "_payment":
            return super().__setattr__(key, value)
        try:
            data = json.loads(self._payment.extra_data)
        except ValueError:
            data = {}
        data[key] = value
        self._payment.extra_data = json.dumps(data)


class BasePayment(models.Model):
    """
    Represents a single transaction. Each instance has one or more PaymentItem.
    """

    variant = models.CharField(max_length=255)
    #: Transaction status
    status = models.CharField(
        max_length=10, choices=PaymentStatus.CHOICES, default=PaymentStatus.WAITING
    )
    fraud_status = models.CharField(
        _("fraud check"),
        max_length=10,
        choices=FraudStatus.CHOICES,
        default=FraudStatus.UNKNOWN,
    )
    fraud_message = models.TextField(blank=True, default="")
    #: Creation date and time
    created = models.DateTimeField(auto_now_add=True)
    #: Date and time of last modification
    modified = models.DateTimeField(auto_now=True)
    #: Transaction ID (if applicable)
    transaction_id = models.CharField(max_length=255, blank=True)
    #: Currency code (may be provider-specific)
    currency = models.CharField(max_length=10)
    #: Total amount (gross)
    total = models.DecimalField(max_digits=9, decimal_places=2, default="0.0")
    delivery = models.DecimalField(max_digits=9, decimal_places=2, default="0.0")
    tax = models.DecimalField(max_digits=9, decimal_places=2, default="0.0")
    description = models.TextField(blank=True, default="")
    billing_first_name = models.CharField(max_length=256, blank=True)
    billing_last_name = models.CharField(max_length=256, blank=True)
    billing_address_1 = models.CharField(max_length=256, blank=True)
    billing_address_2 = models.CharField(max_length=256, blank=True)
    billing_city = models.CharField(max_length=256, blank=True)
    billing_postcode = models.CharField(max_length=256, blank=True)
    billing_country_code = models.CharField(max_length=2, blank=True)
    billing_country_area = models.CharField(max_length=256, blank=True)
    billing_email = models.EmailField(blank=True)
    billing_phone = PhoneNumberField(blank=True)
    customer_ip_address = models.GenericIPAddressField(blank=True, null=True)
    extra_data = models.TextField(blank=True, default="")
    message = models.TextField(blank=True, default="")
    token = models.CharField(max_length=36, blank=True, default="")
    captured_amount = models.DecimalField(max_digits=9, decimal_places=2, default="0.0")

    class Meta:
        abstract = True

    def change_status(self, status: Union[PaymentStatus, str], message=""):
        """
        Updates the Payment status and sends the status_changed signal.
        """
        from .signals import status_changed

        self.status = status
        self.message = message
        self.save(update_fields=["status", "message"])
        status_changed.send(sender=type(self), instance=self)

    def change_fraud_status(self, status: PaymentStatus, message="", commit=True):
        available_statuses = [choice[0] for choice in FraudStatus.CHOICES]
        if status not in available_statuses:
            raise ValueError(
                'Wrong status "{}", it should be one of: {}'.format(
                    status, ", ".join(available_statuses)
                )
            )
        self.fraud_status = status
        self.fraud_message = message
        if commit:
            self.save()

    def save(self, **kwargs):
        if not self.token:
            tries = {}  # Stores a set of tried values
            while True:
                token = str(uuid4())
                if (
                    token in tries and len(tries) >= 100
                ):  # After 100 tries we are impliying an infinite loop
                    raise SystemExit("A possible infinite loop was detected")
                else:
                    if not self.__class__._default_manager.filter(token=token).exists():
                        self.token = token
                        break
                tries.add(token)

        return super().save(**kwargs)

    def __str__(self):
        return self.variant

    def get_form(self, data=None):
        """Return a form to be rendered to complete this payment.

        Please note that this may raise a :class:`~.RedirectNeeded` exception. In this
        case, the user should be redirected to the supplied URL.

        Note that not all providers support a pure form-based flow; some will
        immediately raise ``RedirectNeeded``.
        """
        provider = provider_factory(self.variant, self)
        return provider.get_form(self, data=data)

    def get_purchased_items(self) -> Iterable[PurchasedItem]:
        """Return an iterable of purchased items.

        This information is sent to the payment processor when initiating the payment
        flow. See :class:`.PurchasedItem` for details.

        Subclasses MUST implement this method.
        """

        return []

    def get_failure_url(self) -> str:
        """URL where users will be redirected after a failed payment.

        Return the URL where users will be redirected after a failed attempt to complete a
        payment. This is usually a page explaining the situation to the user with an
        option to retry the payment.

        Note that the URL may contain the ID of this payment, allowing
        the target page to show relevant contextual information.

        Subclasses MUST implement this method.
        """
        raise NotImplementedError()

    def get_success_url(self) -> str:
        """URL where users will be redirected after a successful payment.

        Return the URL where users will be redirected after a successful payment. This
        is usually a page showing a payment summary, though it's application-dependant
        what to show on it.

        Note that the URL may contain the ID of this payment, allowing
        the target page to show relevant contextual information.

        Subclasses MUST implement this method.
        """
        raise NotImplementedError()

    def get_process_url(self) -> str:
        return reverse("process_payment", kwargs={"token": self.token})

    def capture(self, amount=None):
        """Capture a pre-authorized payment.

        Note that not all providers support this method.
        """
        if self.status != PaymentStatus.PREAUTH:
            raise ValueError("Only pre-authorized payments can be captured.")
        provider = provider_factory(self.variant, self)
        amount = provider.capture(self, amount)
        if amount:
            self.captured_amount = amount
            self.change_status(PaymentStatus.CONFIRMED)

    def release(self):
        """Release a pre-authorized payment.

        Note that not all providers support this method.
        """
        if self.status != PaymentStatus.PREAUTH:
            raise ValueError("Only pre-authorized payments can be released.")
        provider = provider_factory(self.variant, self)
        provider.release(self)
        self.change_status(PaymentStatus.REFUNDED)

    def refund(self, amount=None):
        if self.status != PaymentStatus.CONFIRMED:
            raise ValueError("Only charged payments can be refunded.")
        if amount:
            if amount > self.captured_amount:
                raise ValueError(
                    "Refund amount can not be greater then captured amount"
                )
            provider = provider_factory(self.variant, self)
            amount = provider.refund(self, amount)
            self.captured_amount -= amount
        if self.captured_amount == 0 and self.status != PaymentStatus.REFUNDED:
            self.change_status(PaymentStatus.REFUNDED)
        self.save()

    @property
    def attrs(self):
        """A JSON-serialised wrapper around `extra_data`.

        This property exposes a a dict or list which is serialised into the `extra_data`
        text field. Usage of this wrapper is preferred over accessing the underlying
        field directly.

        You may think of this as a `JSONField` which is saved to the `extra_data`
        column.
        """
        # TODO: Deprecate in favour of JSONField when we drop support for django 2.2.
        return PaymentAttributeProxy(self)



import re
from calendar import monthrange
from datetime import date

from django import forms
from django.core import validators
from django.utils.translation import gettext_lazy as _

from .core import get_credit_card_issuer
from .utils import get_month_choices
from .utils import get_year_choices
from .widgets import CreditCardExpiryWidget
from .widgets import CreditCardNumberWidget


class CreditCardNumberField(forms.CharField):
    widget = CreditCardNumberWidget(
        attrs={"autocomplete": "cc-number", "required": "required"}
    )
    default_error_messages = {
        "invalid": _("Please enter a valid card number"),
        "invalid_type": _("We accept only %(valid_types)s"),
    }

    def __init__(self, valid_types=None, *args, **kwargs):
        self.valid_types = valid_types
        kwargs["max_length"] = kwargs.pop("max_length", 32)
        super().__init__(*args, **kwargs)

    def to_python(self, value):
        if value is not None:
            value = re.sub(r"[\s-]+", "", value)
        return super().to_python(value)

    def validate(self, value):
        card_type, issuer_name = get_credit_card_issuer(value)
        if value in validators.EMPTY_VALUES and self.required:
            raise forms.ValidationError(self.error_messages["required"])
        if value and not self.cart_number_checksum_validation(self, value):
            raise forms.ValidationError(self.error_messages["invalid"])
        if value and self.valid_types is not None and card_type not in self.valid_types:
            valid_types = map(issuer_name, self.valid_types)
            error_message = self.error_messages["invalid_type"] % {
                "valid_types": ", ".join(valid_types)
            }
            raise forms.ValidationError(error_message)

    @staticmethod
    def cart_number_checksum_validation(cls, number):
        digits = []
        even = False
        if not number.isdigit():
            return False
        for digit in reversed(number):
            digit = ord(digit) - ord("0")
            if even:
                digit *= 2
                if digit >= 10:
                    digit = digit % 10 + digit // 10
            digits.append(digit)
            even = not even
        return sum(digits) % 10 == 0 if digits else False


class CreditCardExpiryField(forms.MultiValueField):
    default_error_messages = {
        "invalid_month": "Enter a valid month.",
        "invalid_year": "Enter a valid year.",
    }

    def __init__(self, *args, **kwargs):
        errors = self.default_error_messages.copy()
        if "error_messages" in kwargs:
            errors.update(kwargs["error_messages"])

        fields = (
            forms.ChoiceField(
                choices=get_month_choices(),
                error_messages={"invalid": errors["invalid_month"]},
                widget=forms.Select(
                    attrs={"autocomplete": "cc-exp-month", "required": "required"}
                ),
            ),
            forms.ChoiceField(
                choices=get_year_choices(),
                error_messages={"invalid": errors["invalid_year"]},
                widget=forms.Select(
                    attrs={"autocomplete": "cc-exp-year", "required": "required"}
                ),
            ),
        )

        super().__init__(fields, *args, **kwargs)
        self.widget = CreditCardExpiryWidget(
            widgets=[fields[0].widget, fields[1].widget]
        )

    def clean(self, value):
        exp = super().clean(value)
        if exp and date.today() > exp:
            raise forms.ValidationError(
                "The expiration date you entered is in the past."
            )
        return exp

    def compress(self, data_list):
        if data_list:
            if data_list[1] in validators.EMPTY_VALUES:
                error = self.error_messages["invalid_year"]
                raise forms.ValidationError(error)
            if data_list[0] in validators.EMPTY_VALUES:
                error = self.error_messages["invalid_month"]
                raise forms.ValidationError(error)
            year = int(data_list[1])
            month = int(data_list[0])
            # find last day of the month
            day = monthrange(year, month)[1]
            return date(year, month, day)
        return None


class CreditCardVerificationField(forms.CharField):
    widget = forms.TextInput(attrs={"autocomplete": "cc-csc"})
    default_error_messages = {"invalid": _("Enter a valid security number.")}

    def __init__(self, *args, **kwargs):
        kwargs["max_length"] = kwargs.pop("max_length", 4)
        super().__init__(*args, **kwargs)

    def validate(self, value):
        if value in validators.EMPTY_VALUES and self.required:
            raise forms.ValidationError(self.error_messages["required"])
        if value and not re.match("^[0-9]{3,4}$", value):
            raise forms.ValidationError(self.error_messages["invalid"])


class CreditCardNameField(forms.CharField):
    widget = forms.TextInput(attrs={"autocomplete": "cc-name", "required": "required"})



import json
import logging
import re
from uuid import uuid4

from django.http import HttpRequest
from django.http import HttpResponse
from django.shortcuts import redirect
from mercadopago import SDK

from payments import PaymentError
from payments import PaymentStatus
from payments import RedirectNeeded
from payments.core import BasicProvider
from payments.models import BasePayment

logger = logging.getLogger(__name__)

STATUS_MAP = {
    "pending": PaymentStatus.WAITING,
    "approved": PaymentStatus.CONFIRMED,
    "authorized": PaymentStatus.PREAUTH,
    "in_process": PaymentStatus.WAITING,
    "in_mediation": PaymentStatus.WAITING,
    "rejected": PaymentStatus.REJECTED,
    "cancelled": PaymentStatus.ERROR,
    "refunded": PaymentStatus.REFUNDED,
    "charged_back": PaymentStatus.REFUNDED,
}


class MercadoPagoProvider(BasicProvider):
    """This backend implements payments using `MercadoPago <https://www.mercadopago.com.ar/>`_.

    You'll need to install with extra dependencies to use this::

        pip install "django-payments[mercadopago]"

    :param access_token: The access token provided by MP.
    :param sandbox: Whether to use sandbox more.
    """

    def __init__(self, access_token: str, sandbox: bool):
        # self._capture = True
        self.client = SDK(access_token)
        self.is_sandbox = sandbox

    def get_or_create_preference(self, payment: BasePayment):
        if payment.transaction_id:
            return self.get_preference(payment)
        else:
            return self.create_preference(payment)

    def get_preference(self, payment: BasePayment):
        """Helper to fetch the preference for a payment."""
        if not payment.transaction_id:
            raise ValueError("This payment does not have a preference.")

        result = self.client.preference().get(payment.transaction_id)

        if result["status"] >= 300:
            raise PaymentError(
                message="Failed to retrieve MercadoPago preference.",
                code=result["status"],
                gateway_message=result["response"],
            )

        return result["response"]

    def create_preference(self, payment: BasePayment):
        if payment.transaction_id:
            raise ValueError("This payment already has a preference.")

        payment.attrs.external_reference = uuid4().hex

        payload = {
            "auto_return": "all",
            "items": [
                {
                    # "category_id": "services",
                    "currency_id": item.currency,
                    "description": item.sku,
                    "quantity": item.quantity,
                    "title": item.name,
                    "unit_price": float(item.price),
                }
                for item in payment.get_purchased_items()
            ],
            "external_reference": payment.attrs.external_reference,
            "back_urls": {
                "success": self.get_return_url(payment),
                "pending": self.get_return_url(payment),
                "failure": self.get_return_url(payment),
            },
            "notification_url": self.get_return_url(payment),
            "statement_descriptor": payment.description,
        }
        # Payment objects can implement "get_shipment" to use MercadoPago's
        # shipping service.
        if hasattr(payment, "get_shipment"):
            shipments = payment.get_shipment()
            if shipments:
                payload["shipments"] = shipments

        logger.debug("Creating preference with payload: %s", payload)
        result = self.client.preference().create(payload)

        if result["status"] >= 300:
            raise PaymentError(
                message="Failed to create MercadoPago preference.",
                code=result["status"],
                gateway_message=result["response"],
            )

        payment.transaction_id = result["response"]["id"]
        payment.save()

        return result["response"]

    def get_action(self, payment: BasePayment):
        # This is the form-action. But we don't use a form.
        raise NotImplementedError()

    def process_notification(self, payment: BasePayment, request: HttpRequest):
        data = json.loads(request.body)

        logger.debug(
            "Got notification from mercadopago for %s, params: %s, body: %s.",
            payment.pk,
            request.GET,
            data,
        )

        topic = data.get("topic", None)
        resource = data.get("resource", None)

        if topic == "payment":
            match = re.search(r"(\d+)", resource)
            if not match:
                raise ValueError("Missing resource id in notification.")
            collection_id = match.groups()[0]
            self.process_collection(payment, collection_id)

        return HttpResponse("Thanks")

    def process_callback(self, payment: BasePayment, request: HttpRequest):
        collection_id = request.GET.get("collection_id", None)
        if not collection_id or not collection_id.isdigit():
            payment.change_status(PaymentStatus.ERROR)
            return redirect(payment.get_failure_url())

        self.process_collection(payment, collection_id)

        return redirect(payment.get_success_url())

    def process_collection(self, payment: BasePayment, collection_id):
        """Process a collection event.

        MercadoPago sends us an event when they collect money. This fetches
        details for that and updates the payment accordintly.

        :param colection_id: The collection_id we got from MercadoPago.
        """
        response = self.client.payment().get(collection_id)
        if response["status"] != 200:
            message = "MercadoPago sent invalid payment data."
            # Maybe if it's previously approved keep it that way?
            payment.change_status(PaymentStatus.ERROR, message)

            message = f"{message}: {response}"
            raise PaymentError(message)

        upstream_status = response["response"]["status"]
        payment.change_status(STATUS_MAP[upstream_status])

    def process_data(self, payment: BasePayment, request: HttpRequest):
        """Handle a request received after a payment.

        If it's a GET request, then it's the user being redirected after
        completing a payment (it may have failed or have been successfull).

        If it's a POST, it's a webhook notification.
        """
        if request.method == "GET":
            return self.process_callback(payment, request)
        elif request.method == "POST":
            return self.process_notification(payment, request)

    def get_form(self, payment: BasePayment, data=None):
        # There's no form for MP, we need to redirect to their checkout page.
        preference = self.get_or_create_preference(payment)
        logger.debug("Got preference: %s", preference)

        if self.is_sandbox:
            url = preference["sandbox_init_point"]
        else:
            url = preference["init_point"]

        raise RedirectNeeded(url)

    def capture(self, payment: BasePayment, amount=None):
        # only allow if its PRE_AUTH
        raise NotImplementedError()

    def refund(self, payment: BasePayment, amount=None):
        raise NotImplementedError()

    def poll_for_updates(self, payment: BasePayment):
        """Helper method to fetch any updates on MercadoPago.

        There's occasional times of flakiness with their notification service,
        and this helper method helps recover from that and pick up any missed
        payments.
        """
        data = self.client.payment().search(
            {
                "external_reference": payment.attrs.external_reference,
            }
        )

        logger.info("Found payment info for %s: %s.", payment, data)

        if data["results"]:
            self.process_collection(payment, data["results"][-1]["id"])



from typing import Optional
from unittest.mock import Mock
from unittest.mock import call
from unittest.mock import patch

import pytest
from django.test import RequestFactory

from payments import PaymentError
from payments import PaymentStatus
from payments import PurchasedItem
from payments import RedirectNeeded
from payments.mercadopago import MercadoPagoProvider


class Payment(Mock):
    id = 1
    description = "payment"
    currency = "ARS"
    delivery = 10
    status = PaymentStatus.WAITING
    message = None
    tax = 10
    total = 100
    captured_amount = 0
    transaction_id: Optional[str] = None
    billing_email = "john@doe.com"

    def change_status(self, status, message=""):
        self.status = status
        self.message = message

    def change_fraud_status(self, status, message="", commit=True):
        self.fraud_status = status
        self.fraud_message = message

    def capture(self, amount=None):
        amount = amount or self.total
        self.captured_amount = amount
        self.change_status(PaymentStatus.CONFIRMED)

    def get_success_url(self):
        return "http://example.com/success"

    def get_failure_url(self):
        return "http://example.com/failure"

    def get_process_url(self):
        return "http://example.com/process"

    def get_purchased_items(self):
        yield PurchasedItem(
            name="Some Product",
            sku="SP12",
            quantity=1,
            price="1999",
            currency="ARS",
        )


@pytest.fixture
def mp_provider():
    return MercadoPagoProvider("fake_access_token", True)


# These are mock request and responses are based on valid data that actually worked in
# the live environment. Values have been replaced with random numbers / names /
# etc. However, the structure of the payloads matches what the API currently uses.


def test_approved_payment_notification(rf, mp_provider: MercadoPagoProvider):
    payment = Payment()

    request = rf.get(
        "/payments/process/d16695e8-b76d-4438-bd10-634545ecb1d6/",
        {
            "collection_id": "14543542354",
            "collection_status": "approved",
            "external_reference": "e312623454323b6a8605de11873cbb82",
            "merchant_account_id": "null",
            "merchant_order_id": "2344532517",
            "payment_id": "14345344918",
            "payment_type": "credit_card",
            "preference_id": "151235340-fc9fcecd-6e38-4e50-adac-a10abab4f138",
            "processing_mode": "aggregator",
            "site_id": "MLA",
            "status": "approved",
        },
    )

    payment_info_response = {
        "response": {
            "acquirer_reconciliation": [],
            "additional_info": {
                "available_balance": None,
                "ip_address": "127.19.22.46",
                "items": [
                    {
                        "category_id": None,
                        "description": "454716",
                        "id": None,
                        "picture_url": None,
                        "quantity": "1",
                        "title": "Order #454416",
                        "unit_price": "655.0",
                    }
                ],
                "nsu_processadora": None,
            },
            "authorization_code": "345385",
            "binary_mode": False,
            "brand_id": None,
            "call_for_authorize_id": None,
            "captured": True,
            "card": {
                "cardholder": {
                    "identification": {"number": "45354441", "type": "DNI"},
                    "name": "María Perez",
                },
                "date_created": "2020-10-10T17:27:45.000-04:00",
                "date_last_updated": "2021-04-21T13:37:38.000-04:00",
                "expiration_month": 3,
                "expiration_year": 2023,
                "first_six_digits": "235237",
                "id": "8945698155",
                "last_four_digits": "5847",
            },
            "charges_details": [],
            "collector_id": 153453450,
            "corporation_id": None,
            "counter_currency": None,
            "coupon_amount": 0,
            "currency_id": "ARS",
            "date_approved": "2021-04-21T13:44:17.000-04:00",
            "date_created": "2021-04-21T13:44:15.000-04:00",
            "date_last_updated": "2021-04-21T13:44:17.000-04:00",
            "date_of_expiration": None,
            "deduction_schema": None,
            "description": "Order #343716",
            "differential_pricing_id": None,
            "external_reference": "e2353245324a4b6a8605d565573cbb82",
            "fee_details": [
                {"amount": 47.49, "fee_payer": "collector", "type": "mercadopago_fee"}
            ],
            "id": 34534444918,
            "installments": 1,
            "integrator_id": None,
            "issuer_id": "688",
            "live_mode": True,
            "marketplace_owner": None,
            "merchant_account_id": None,
            "merchant_number": None,
            "metadata": {},
            "money_release_date": "2021-04-21T13:44:17.000-04:00",
            "money_release_schema": None,
            "notification_url": "https://www.example.com/payments/process/d14343e8-b76d-4438-bd10-633453ecb1d6/",
            "operation_type": "regular_payment",
            "order": {"id": "3453449517", "type": "mercadopago"},
            "payer": {
                "email": "mariaperez@example.com",
                "entity_type": None,
                "first_name": "maria",
                "id": "53474545",
                "identification": {"number": "23423341", "type": "DNI"},
                "last_name": "perez",
                "operator_id": None,
                "phone": {"area_code": "011", "extension": None, "number": "40004444"},
                "type": "registered",
            },
            "payment_method_id": "cabal",
            "payment_type_id": "credit_card",
            "platform_id": None,
            "point_of_interaction": {},
            "pos_id": None,
            "processing_mode": "aggregator",
            "refunds": [],
            "shipping_amount": 0,
            "sponsor_id": None,
            "statement_descriptor": "MERCADOPAGO - TEST",
            "status": "approved",
            "status_detail": "accredited",
            "store_id": None,
            "taxes_amount": 0,
            "transaction_amount": 655,
            "transaction_amount_refunded": 0,
            "transaction_details": {
                "acquirer_reference": None,
                "external_resource_url": None,
                "financial_institution": None,
                "installment_amount": 655,
                "net_received_amount": 606.2,
                "overpaid_amount": 0,
                "payable_deferral_period": None,
                "payment_method_reference_id": "423423523525349",
                "total_paid_amount": 655,
            },
        },
        "status": 200,
    }

    with patch(
        "mercadopago.resources.payment.Payment.get",
        spec=True,
        return_value=payment_info_response,
    ) as payment_info, patch(
        "payments.mercadopago.redirect",
        spec=True,
    ) as redirect:
        rv = mp_provider.process_data(payment, request)

    assert payment_info.call_count == 1
    assert rv == redirect(payment.get_success_url())


def test_create_preference_that_already_exists(mp_provider: MercadoPagoProvider):
    payment = Payment()
    payment.transaction_id = "123"

    with pytest.raises(ValueError, match="payment already has a preference"):
        mp_provider.create_preference(payment)


def test_create_preference_failure(mp_provider: MercadoPagoProvider):
    preference_info = {
        "status": 500,
        "response": "internal server error",
    }

    payment = Payment()

    with patch(
        "mercadopago.resources.preference.Preference.create",
        spec=True,
        return_value=preference_info,
    ), pytest.raises(
        PaymentError,
        match="Failed to create MercadoPago preference.",
    ):
        mp_provider.create_preference(payment)


def test_process_successful_collection(mp_provider: MercadoPagoProvider):
    payment_info = {
        "status": 200,
        "response": {"status": "approved"},
    }

    payment = Payment()
    with patch(
        "mercadopago.resources.payment.Payment.get",
        spec=True,
        return_value=payment_info,
    ):
        mp_provider.process_collection(payment, "12")

    assert payment.status == PaymentStatus.CONFIRMED


def test_process_failed_collection(mp_provider: MercadoPagoProvider):
    payment_info = {
        "status": 404,
    }

    payment = Payment()
    with patch(
        "mercadopago.resources.payment.Payment.get",
        spec=True,
        return_value=payment_info,
    ), pytest.raises(
        PaymentError,
        match="MercadoPago sent invalid payment data.",
    ):
        mp_provider.process_collection(payment, "12")

    assert payment.status == PaymentStatus.ERROR


def test_process_pending_collection(mp_provider: MercadoPagoProvider):
    payment_info = {
        "status": 200,
        "response": {"status": "pending"},
    }

    payment = Payment()
    with patch(
        "mercadopago.resources.payment.Payment.get",
        spec=True,
        return_value=payment_info,
    ):
        mp_provider.process_collection(payment, "12")

    assert payment.status == PaymentStatus.WAITING


def test_get_preference(mp_provider: MercadoPagoProvider):
    preference = Mock()
    mocked_response = {
        "status": 200,
        "response": preference,
    }

    payment = Payment()
    payment.transaction_id = "ABJ122"
    with patch(
        "mercadopago.resources.preference.Preference.get",
        spec=True,
        return_value=mocked_response,
    ) as get_preference:
        assert mp_provider.get_preference(payment) == preference

    assert get_preference.call_count == 1
    assert get_preference.call_args == call(payment.transaction_id)


def test_get_preference_with_missing_transaction_id(mp_provider: MercadoPagoProvider):
    payment = Payment()

    with pytest.raises(ValueError, match="payment does not have a preference"):
        mp_provider.get_preference(payment)


def test_get_preference_internal_error(mp_provider: MercadoPagoProvider):
    mocked_response = {
        "status": 500,
        "response": "Internal error",
    }

    payment = Payment()
    payment.transaction_id = "ABJ122"
    with patch(
        "mercadopago.resources.preference.Preference.get",
        spec=True,
        return_value=mocked_response,
    ) as get_preference:
        with pytest.raises(
            PaymentError, match="Failed to retrieve MercadoPago preference."
        ):
            mp_provider.get_preference(payment)

    assert get_preference.call_count == 1
    assert get_preference.call_args == call(payment.transaction_id)


@pytest.mark.parametrize(
    "is_sandbox,url_attr",
    [
        (False, "init_point"),
        (True, "sandbox_init_point"),
    ],
)
def test_get_form_for_existing_preference(
    mp_provider: MercadoPagoProvider,
    is_sandbox: bool,
    url_attr: str,
):
    mp_provider.is_sandbox = is_sandbox

    mocked_response = {
        "status": 200,
        "response": {url_attr: "https://example.com/pay"},
    }

    payment = Payment()
    payment.transaction_id = "ABJ122"
    with patch(
        "mercadopago.resources.preference.Preference.get",
        spec=True,
        return_value=mocked_response,
    ) as get_preference:
        with pytest.raises(RedirectNeeded) as exc_info:
            mp_provider.get_form(payment)

    assert get_preference.call_count == 1
    assert get_preference.call_args == call(payment.transaction_id)
    assert str(exc_info.value) == "https://example.com/pay"


def test_get_form_for_inexistent_preference(mp_provider: MercadoPagoProvider):
    mocked_response = {
        "status": 200,
        "response": {
            "id": "AZ12",
            "sandbox_init_point": "https://example.com/pay",
        },
    }

    payment = Payment()
    with patch(
        "mercadopago.resources.preference.Preference.create",
        spec=True,
        return_value=mocked_response,
    ) as get_preference:
        with pytest.raises(RedirectNeeded) as exc_info:
            mp_provider.get_form(payment)

    assert get_preference.call_count == 1
    assert str(exc_info.value) == "https://example.com/pay"


def test_process_notification_ignores_merchant_orders(
    mp_provider: MercadoPagoProvider,
    rf: RequestFactory,
):
    payment = Payment()
    request = rf.post(
        "/process/xxx-yyy",
        {"topic": "merchant_order"},
        content_type="application/json",
    )
    response = mp_provider.process_data(payment, request)

    assert response.content.decode() == "Thanks"


def test_process_notification_processes_payment_collection(
    mp_provider: MercadoPagoProvider,
    rf: RequestFactory,
):
    payment = Payment()
    request = rf.post(
        "/process/xxx-yyy",
        {"topic": "payment", "resource": "payment123"},
        content_type="application/json",
    )
    with patch(
        "payments.mercadopago.MercadoPagoProvider.process_collection",
        spec=True,
    ) as process_collection:
        response = mp_provider.process_data(payment, request)

    assert response.content.decode() == "Thanks"
    assert process_collection.call_count == 1
    assert process_collection.call_args == call(payment, "123")



import json
import logging
from datetime import timedelta
from decimal import ROUND_HALF_UP
from decimal import Decimal
from functools import wraps

import requests
from django.http import HttpResponseForbidden
from django.shortcuts import redirect
from django.utils import timezone
from requests.exceptions import HTTPError

from .. import PaymentError
from .. import PaymentStatus
from .. import RedirectNeeded
from ..core import BasicProvider
from ..core import get_credit_card_issuer
from .forms import PaymentForm

# Get an instance of a logger
logger = logging.getLogger(__name__)

CENTS = Decimal("0.01")


class UnauthorizedRequest(Exception):
    pass


def authorize(fun):
    @wraps(fun)
    def wrapper(*args, **kwargs):
        self = args[0]
        payment = args[1]
        self.access_token = self.get_access_token(payment)
        try:
            response = fun(*args, **kwargs)
        except HTTPError as e:
            if e.response.status_code == 401:
                last_auth_response = self.get_last_response(payment, is_auth=True)
                if "access_token" in last_auth_response:
                    del last_auth_response["access_token"]
                    self.set_response_data(payment, last_auth_response, is_auth=True)
                self.access_token = self.get_access_token(payment)
                response = fun(*args, **kwargs)
            else:
                raise
        return response

    return wrapper


class PaypalProvider(BasicProvider):
    """Payment provider for Paypal, redirection-based.

    This backend implements payments using `PayPal.com <https://www.paypal.com/>`_.

    :param client_id: Client ID assigned by PayPal or your email address
    :param secret: Secret assigned by PayPal
    :param endpoint: The API endpoint to use. For the production environment, use ``'https://api.paypal.com'`` instead
    :param capture: Whether to capture the payment automatically. See :ref:`capture-payments` for more details.
    """

    def __init__(
        self, client_id, secret, endpoint="https://api.sandbox.paypal.com", capture=True
    ):
        self.secret = secret
        self.client_id = client_id
        self.endpoint = endpoint
        self.oauth2_url = self.endpoint + "/v1/oauth2/token"
        self.payments_url = self.endpoint + "/v1/payments/payment"
        self.payment_execute_url = self.payments_url + "/%(id)s/execute/"
        self.payment_refund_url = (
            self.endpoint + "/v1/payments/capture/{captureId}/refund"
        )
        super().__init__(capture=capture)

    def set_response_data(self, payment, response, is_auth=False):
        extra_data = json.loads(payment.extra_data or "{}")
        if is_auth:
            extra_data["auth_response"] = response
        else:
            extra_data["response"] = response
            if "links" in response:
                extra_data["links"] = {link["rel"]: link for link in response["links"]}
        payment.extra_data = json.dumps(extra_data)
        payment.save()

    def set_response_links(self, payment, response):
        transaction = response["transactions"][0]
        related_resources = transaction["related_resources"][0]
        resource_key = "sale" if self._capture else "authorization"
        links = related_resources[resource_key]["links"]
        extra_data = json.loads(payment.extra_data or "{}")
        extra_data["links"] = {link["rel"]: link for link in links}
        payment.extra_data = json.dumps(extra_data)
        payment.save()

    def set_error_data(self, payment, error):
        extra_data = json.loads(payment.extra_data or "{}")
        extra_data["error"] = error
        payment.extra_data = json.dumps(extra_data)
        payment.save()

    def _get_links(self, payment):
        extra_data = json.loads(payment.extra_data or "{}")
        links = extra_data.get("links", {})
        return links

    @authorize
    def post(self, payment, *args, **kwargs):
        kwargs["headers"] = {
            "Content-Type": "application/json",
            "Authorization": self.access_token,
        }
        if "data" in kwargs:
            kwargs["data"] = json.dumps(kwargs["data"])
        response = requests.post(*args, **kwargs)
        try:
            data = response.json()
        except ValueError:
            data = {}
        if 400 <= response.status_code <= 500:
            self.set_error_data(payment, data)
            logger.debug(data)
            message = "Paypal error"
            if response.status_code == 400:
                error_data = response.json()
                logger.warning(
                    message,
                    extra={"response": error_data, "status_code": response.status_code},
                )
                message = error_data.get("message", message)
            else:
                logger.warning(message, extra={"status_code": response.status_code})
            payment.change_status(PaymentStatus.ERROR, message)
            raise PaymentError(message)
        else:
            self.set_response_data(payment, data)
        return data

    def get_last_response(self, payment, is_auth=False):
        extra_data = json.loads(payment.extra_data or "{}")
        if is_auth:
            return extra_data.get("auth_response", {})
        return extra_data.get("response", {})

    def get_access_token(self, payment):
        last_auth_response = self.get_last_response(payment, is_auth=True)
        created = payment.created
        now = timezone.now()
        if (
            "access_token" in last_auth_response
            and "expires_in" in last_auth_response
            and (created + timedelta(seconds=last_auth_response["expires_in"])) > now
        ):
            return "{} {}".format(
                last_auth_response["token_type"], last_auth_response["access_token"]
            )
        else:
            headers = {"Accept": "application/json", "Accept-Language": "en_US"}
            post = {"grant_type": "client_credentials"}
            response = requests.post(
                self.oauth2_url,
                data=post,
                headers=headers,
                auth=(self.client_id, self.secret),
            )
            response.raise_for_status()
            data = response.json()
            last_auth_response.update(data)
            self.set_response_data(payment, last_auth_response, is_auth=True)
            return "{} {}".format(data["token_type"], data["access_token"])

    def get_transactions_items(self, payment):
        for purchased_item in payment.get_purchased_items():
            price = purchased_item.price.quantize(CENTS, rounding=ROUND_HALF_UP)
            item = {
                "name": purchased_item.name[:127],
                "quantity": str(purchased_item.quantity),
                "price": str(price),
                "currency": purchased_item.currency,
                "sku": purchased_item.sku,
            }
            yield item

    def get_transactions_data(self, payment):
        items = list(self.get_transactions_items(payment))
        sub_total = payment.total - payment.delivery - payment.tax
        sub_total = sub_total.quantize(CENTS, rounding=ROUND_HALF_UP)
        total = payment.total.quantize(CENTS, rounding=ROUND_HALF_UP)
        tax = payment.tax.quantize(CENTS, rounding=ROUND_HALF_UP)
        delivery = payment.delivery.quantize(CENTS, rounding=ROUND_HALF_UP)
        data = {
            "intent": "sale" if self._capture else "authorize",
            "transactions": [
                {
                    "amount": {
                        "total": str(total),
                        "currency": payment.currency,
                        "details": {
                            "subtotal": str(sub_total),
                            "tax": str(tax),
                            "shipping": str(delivery),
                        },
                    },
                    "item_list": {"items": items},
                    "description": payment.description,
                }
            ],
        }
        return data

    def get_product_data(self, payment, extra_data=None):
        return_url = self.get_return_url(payment)
        data = self.get_transactions_data(payment)
        data["redirect_urls"] = {"return_url": return_url, "cancel_url": return_url}
        data["payer"] = {"payment_method": "paypal"}
        return data

    def get_form(self, payment, data=None):
        if not payment.id:
            payment.save()
        links = self._get_links(payment)
        redirect_to = links.get("approval_url")
        if not redirect_to:
            payment_data = self.create_payment(payment)
            payment.transaction_id = payment_data["id"]
            links = self._get_links(payment)
            redirect_to = links["approval_url"]
        payment.change_status(PaymentStatus.WAITING)
        raise RedirectNeeded(redirect_to["href"])

    def process_data(self, payment, request):
        success_url = payment.get_success_url()
        failure_url = payment.get_failure_url()
        if "token" not in request.GET:
            return HttpResponseForbidden("FAILED")
        payer_id = request.GET.get("PayerID")
        if not payer_id:
            if payment.status != PaymentStatus.CONFIRMED:
                payment.change_status(PaymentStatus.REJECTED)
                return redirect(failure_url)
            else:
                return redirect(success_url)
        try:
            executed_payment = self.execute_payment(payment, payer_id)
        except PaymentError:
            return redirect(failure_url)
        self.set_response_links(payment, executed_payment)
        payment.attrs.payer_info = executed_payment["payer"]["payer_info"]
        if self._capture:
            payment.captured_amount = payment.total
            payment.change_status(PaymentStatus.CONFIRMED)
        else:
            payment.change_status(PaymentStatus.PREAUTH)
        return redirect(success_url)

    def create_payment(self, payment, extra_data=None):
        product_data = self.get_product_data(payment, extra_data)
        payment = self.post(payment, self.payments_url, data=product_data)
        return payment

    def execute_payment(self, payment, payer_id):
        post = {"payer_id": payer_id}
        links = self._get_links(payment)
        execute_url = links["execute"]["href"]
        return self.post(payment, execute_url, data=post)

    def get_amount_data(self, payment, amount=None):
        return {
            "currency": payment.currency,
            "total": str(amount.quantize(CENTS, rounding=ROUND_HALF_UP)),
        }

    def capture(self, payment, amount=None):
        if amount is None:
            amount = payment.total
        amount_data = self.get_amount_data(payment, amount)
        capture_data = {"amount": amount_data, "is_final_capture": True}
        links = self._get_links(payment)
        url = links["capture"]["href"]
        try:
            capture = self.post(payment, url, data=capture_data)
        except HTTPError as e:
            try:
                error = e.response.json()
            except ValueError:
                error = {}
            if error.get("name") != "AUTHORIZATION_ALREADY_COMPLETED":
                raise e
            capture = {"state": "completed"}
        state = capture["state"]
        if state == "completed":
            payment.change_status(PaymentStatus.CONFIRMED)
            return amount
        elif state in ["partially_captured", "partially_refunded"]:
            return amount
        elif state == "pending":
            payment.change_status(PaymentStatus.WAITING)
        elif state == "refunded":
            payment.change_status(PaymentStatus.REFUNDED)
            raise PaymentError("Payment already refunded")

    def release(self, payment):
        links = self._get_links(payment)
        url = links["void"]["href"]
        self.post(payment, url)

    def refund(self, payment, amount=None):
        if amount is None:
            amount = payment.captured_amount
        amount_data = self.get_amount_data(payment, amount)
        refund_data = {"amount": amount_data}
        links = self._get_links(payment)
        url = links["refund"]["href"]
        self.post(payment, url, data=refund_data)
        payment.change_status(PaymentStatus.REFUNDED)
        return amount


class PaypalCardProvider(PaypalProvider):
    """Payment provider for Paypal, form-based.

    This backend implements payments using `PayPal.com <https://www.paypal.com/>`_ but
    the credit card data is collected by your site.

    Parameters are the same as  :class:`~PaypalProvider`.

    This backend does not support fraud detection.
    """

    def get_form(self, payment, data=None):
        if payment.status == PaymentStatus.WAITING:
            payment.change_status(PaymentStatus.INPUT)
        form = PaymentForm(data, provider=self, payment=payment)
        if form.is_valid():
            raise RedirectNeeded(payment.get_success_url())
        return form

    def get_product_data(self, payment, extra_data=None):
        extra_data = extra_data or {}
        data = self.get_transactions_data(payment)
        year = extra_data["expiration"].year
        month = extra_data["expiration"].month
        number = extra_data["number"]
        card_type, _card_issuer = get_credit_card_issuer(number)
        credit_card = {
            "number": number,
            "type": card_type,
            "expire_month": month,
            "expire_year": year,
        }
        if "cvv2" in extra_data and extra_data["cvv2"]:
            credit_card["cvv2"] = extra_data["cvv2"]
        data["payer"] = {
            "payment_method": "credit_card",
            "funding_instruments": [{"credit_card": credit_card}],
        }
        return data

    def process_data(self, payment, request):
        return HttpResponseForbidden("FAILED")



import json
from unittest import TestCase
from unittest.mock import MagicMock
from unittest.mock import Mock
from unittest.mock import patch

from .. import PaymentStatus
from .. import RedirectNeeded
from . import SofortProvider

SECRET = "abcd1234"
CLIENT_ID = "1234"
PROJECT_ID = "abcd"


class Payment(Mock):
    id = 1
    variant = "sagepay"
    currency = "USD"
    total = 100
    status = PaymentStatus.WAITING
    transaction_id = None
    captured_amount = 0
    billing_first_name = "John"
    description = "foo bar"

    def get_process_url(self):
        return "http://example.com"

    def get_failure_url(self):
        return "http://cancel.com"

    def get_success_url(self):
        return "http://success.com"

    def change_status(self, status):
        self.status = status


class TestSofortProvider(TestCase):
    def setUp(self):
        self.payment = Payment()
        self.provider = SofortProvider(id=CLIENT_ID, project_id=PROJECT_ID, key=SECRET)

    @patch("xmltodict.parse")
    @patch("requests.post")
    def test_provider_raises_redirect_needed_on_success(
        self, mocked_post, mocked_parser
    ):
        response = MagicMock()
        response.status_code = 200
        mocked_post.return_value = response
        mocked_parser.return_value = {
            "new_transaction": {"payment_url": "http://payment.com"}
        }
        with self.assertRaises(RedirectNeeded):
            self.provider.get_form(self.payment)

    @patch("xmltodict.parse")
    @patch("requests.post")
    @patch("payments.sofort.redirect")
    def test_provider_redirects_on_success(
        self, mocked_redirect, mocked_post, mocked_parser
    ):
        transaction_id = "1234"
        request = MagicMock()
        request.GET = {"trans": transaction_id}
        mocked_parser.return_value = {
            "transactions": {
                "transaction_details": {
                    "status": "ok",
                    "sender": {"holder": "John Doe", "country_code": "EN"},
                }
            }
        }
        self.provider.process_data(self.payment, request)
        self.assertEqual(self.payment.status, PaymentStatus.CONFIRMED)
        self.assertEqual(self.payment.captured_amount, self.payment.total)
        self.assertEqual(self.payment.transaction_id, transaction_id)

    @patch("xmltodict.parse")
    @patch("requests.post")
    @patch("payments.sofort.redirect")
    def test_provider_redirects_on_failure(
        self, mocked_redirect, mocked_post, mocked_parser
    ):
        transaction_id = "1234"
        request = MagicMock()
        request.GET = {"trans": transaction_id}
        mocked_parser.return_value = {}
        self.provider.process_data(self.payment, request)
        self.assertEqual(self.payment.status, PaymentStatus.REJECTED)
        self.assertEqual(self.payment.captured_amount, 0)
        self.assertEqual(self.payment.transaction_id, transaction_id)

    @patch("xmltodict.parse")
    @patch("requests.post")
    def test_provider_refunds_payment(self, mocked_post, mocked_parser):
        self.payment.extra_data = json.dumps(
            {
                "transactions": {
                    "transaction_details": {
                        "status": "ok",
                        "sender": {
                            "holder": "John Doe",
                            "country_code": "EN",
                            "bic": "1234",
                            "iban": "abcd",
                        },
                    }
                }
            }
        )
        mocked_parser.return_value = {}
        self.provider.refund(self.payment)
        self.assertEqual(self.payment.status, PaymentStatus.REFUNDED)



from unittest import TestCase
from unittest.mock import MagicMock
from unittest.mock import Mock
from unittest.mock import patch

from .. import PaymentStatus
from . import SagepayProvider

VENDOR = "abcd1234"
ENCRYPTION_KEY = "1234abdd1234abcd"


class Payment(Mock):
    id = 1
    variant = "sagepay"
    currency = "USD"
    total = 100
    status = PaymentStatus.WAITING
    transaction_id = None
    captured_amount = 0
    billing_first_name = "John"

    def get_process_url(self):
        return "http://example.com"

    def get_failure_url(self):
        return "http://cancel.com"

    def get_success_url(self):
        return "http://success.com"

    def change_status(self, status):
        self.status = status


class TestSagepayProvider(TestCase):
    def setUp(self):
        self.payment = Payment()
        self.provider = SagepayProvider(vendor=VENDOR, encryption_key=ENCRYPTION_KEY)

    @patch("payments.sagepay.redirect")
    def test_provider_raises_redirect_needed_on_success(self, mocked_redirect):
        data = {"Status": "OK"}
        data = "&".join("%s=%s" % kv for kv in data.items())
        with patch.object(SagepayProvider, "aes_dec", return_value=data):
            self.provider.process_data(self.payment, MagicMock())
            self.assertEqual(self.payment.status, PaymentStatus.CONFIRMED)
            self.assertEqual(self.payment.captured_amount, self.payment.total)

    @patch("payments.sagepay.redirect")
    def test_provider_raises_redirect_needed_on_failure(self, mocked_redirect):
        data = {"Status": ""}
        data = "&".join("%s=%s" % kv for kv in data.items())
        with patch.object(SagepayProvider, "aes_dec", return_value=data):
            self.provider.process_data(self.payment, MagicMock())
            self.assertEqual(self.payment.status, PaymentStatus.REJECTED)
            self.assertEqual(self.payment.captured_amount, 0)

    def test_provider_encrypts_data(self):
        data = self.provider.get_hidden_fields(self.payment)
        decrypted_data = self.provider.aes_dec(data["Crypt"])
        self.assertIn(self.payment.billing_first_name, str(decrypted_data))

    def test_encrypt_method_returns_valid_data(self):
        encrypted = self.provider.aes_enc("mirumee")
        self.assertEqual(encrypted, b"@e63c293672f50b9c8e291831facb4e4f")



import hashlib
from unittest import TestCase
from unittest.mock import MagicMock
from unittest.mock import Mock

from django.http import HttpResponse
from django.http import HttpResponseForbidden

from .. import PaymentStatus
from . import DotpayProvider
from .forms import COMPLETED
from .forms import REJECTED

VARIANT = "dotpay"
PIN = "123"
PROCESS_POST = {
    "id": "123",
    "operation_number": "MI-5",
    "operation_type": "payment",
    "operation_status": COMPLETED,
    "operation_amount": "212.35",
    "operation_currency": "PLN",
    "operation_original_amount": "58.07",
    "operation_original_currency": "USD",
    "operation_datetime": "2018-02-28 12:00:00",
    "control": "1",
    "description": "Order #1",
    "email": "user@example.org",
    "p_info": "John Doe (seller@example.com)",
    "p_email": "seller@example.com",
    "channel": "1",
}


def get_post_with_sha256(post):
    post = post.copy()
    post["pin"] = "123"
    keys = [
        "pin",
        "id",
        "operation_number",
        "operation_type",
        "operation_status",
        "operation_amount",
        "operation_currency",
        "operation_original_amount",
        "operation_original_currency",
        "operation_datetime",
        "control",
        "description",
        "email",
        "p_info",
        "p_email",
        "channel",
    ]
    key = "".join([post[key] for key in keys])
    sha256 = hashlib.sha256()
    sha256.update(key.encode("utf-8"))
    key_hash = sha256.hexdigest()
    post["signature"] = key_hash
    return post


class Payment(Mock):
    id = 1
    variant = VARIANT
    currency = "USD"
    total = 100
    status = PaymentStatus.WAITING

    def get_process_url(self):
        return "http://example.com"

    def get_failure_url(self):
        return "http://cancel.com"

    def get_success_url(self):
        return "http://success.com"

    def change_status(self, status):
        self.status = status


class TestDotpayProvider(TestCase):
    def setUp(self):
        self.payment = Payment()

    def test_get_hidden_fields(self):
        """DotpayProvider.get_hidden_fields() returns a dictionary"""
        provider = DotpayProvider(seller_id="123", pin=PIN)
        self.assertEqual(type(provider.get_hidden_fields(self.payment)), dict)

    def test_process_data_payment_accepted(self):
        """DotpayProvider.process_data() returns a correct HTTP response"""
        request = MagicMock()
        request.POST = get_post_with_sha256(PROCESS_POST)
        params = {
            "seller_id": 123,
            "pin": PIN,
            "endpoint": "test.endpoint.com",
            "channel": 1,
            "lang": "en",
            "lock": True,
        }
        provider = DotpayProvider(**params)
        response = provider.process_data(self.payment, request)
        self.assertEqual(type(response), HttpResponse)
        self.assertEqual(self.payment.status, PaymentStatus.CONFIRMED)

    def test_process_data_payment_rejected(self):
        """DotpayProvider.process_data() returns a correct HTTP response"""
        data = dict(PROCESS_POST)
        data.update({"operation_status": REJECTED})
        request = MagicMock()
        request.POST = get_post_with_sha256(data)
        params = {
            "seller_id": 123,
            "pin": PIN,
            "endpoint": "test.endpoint.com",
            "channel": 1,
            "lang": "en",
            "lock": True,
        }
        provider = DotpayProvider(**params)
        response = provider.process_data(self.payment, request)
        self.assertEqual(type(response), HttpResponse)
        self.assertEqual(self.payment.status, PaymentStatus.REJECTED)

    def test_incorrect_process_data(self):
        """DotpayProvider.process_data() checks POST signature"""
        request = MagicMock()
        request.POST = PROCESS_POST
        provider = DotpayProvider(seller_id="123", pin=PIN)
        response = provider.process_data(self.payment, request)
        self.assertEqual(type(response), HttpResponseForbidden)

    def test_uses_channel_groups_when_set(self):
        channel_groups = "K,T"
        params = {
            "seller_id": 123,
            "pin": PIN,
            "endpoint": "test.endpoint.com",
            "channel": 1,
            "channel_groups": channel_groups,
            "lang": "en",
            "lock": True,
        }
        provider = DotpayProvider(**params)
        hidden_fields = provider.get_hidden_fields(self.payment)
        self.assertEqual(hidden_fields["channel_groups"], channel_groups)
        self.assertIsNone(hidden_fields.get("channel"))



from pathlib import Path

import dj_database_url

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/3.2/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "django-insecure-4cz7$eek%+vryv*!p#+zd!e*xv@*1dtxpxnv(it=r1yys#l554"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = ["localhost", "127.0.0.1"]

# Application definition

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.sites",
    "testapp.testmain",
    "payments",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "testapp.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "testapp.wsgi.application"

# Database
# https://docs.djangoproject.com/en/3.2/ref/settings/#databases

DATABASES = {"default": dj_database_url.config()}

# Password validation
# https://docs.djangoproject.com/en/3.2/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]

# Internationalization
# https://docs.djangoproject.com/en/3.2/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.2/howto/static-files/

STATIC_URL = "/static/"

# Default primary key field type
# https://docs.djangoproject.com/en/3.2/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

PAYMENT_HOST = "localhost:8000"
PAYMENT_USES_SSL = False
PAYMENT_MODEL = "testmain.Payment"
PAYMENT_VARIANTS = {
    "default": ("payments.dummy.DummyProvider", {}),
}



import hashlib
import json
from decimal import Decimal
from unittest import TestCase
from unittest.mock import MagicMock
from unittest.mock import patch

from django.http import HttpResponse
from django.http import HttpResponseForbidden

from .. import PaymentStatus
from . import CoinbaseProvider

PAYMENT_TOKEN = "5a4dae68-2715-4b1e-8bb2-2c2dbe9255f6"
KEY = "abc123"
SECRET = "123abc"
VARIANT = "coinbase"

COINBASE_REQUEST = {
    "order": {
        "transaction": {"id": "123456"},
        "custom": hashlib.md5(
            ("coinbase-{}-{}".format(PAYMENT_TOKEN, KEY)).encode("utf-8")
        ).hexdigest(),
    }
}


class Payment:
    id = 1
    description = "payment"
    currency = "BTC"
    total = Decimal(100)
    status = PaymentStatus.WAITING
    token = PAYMENT_TOKEN
    variant = VARIANT

    def change_status(self, status):
        self.status = status

    def get_failure_url(self):
        return "http://cancel.com"

    def get_process_url(self):
        return "http://example.com"

    def get_purchased_items(self):
        return []

    def save(self):
        return self

    def get_success_url(self):
        return "http://success.com"


class TestCoinbaseProvider(TestCase):
    def setUp(self):
        self.payment = Payment()
        self.provider = CoinbaseProvider(key=KEY, secret=SECRET)

    def test_process_data(self):
        """
        CoinbaseProvider.process_data() returns a correct HTTP response
        """
        request = MagicMock()
        request.body = json.dumps(COINBASE_REQUEST)
        response = self.provider.process_data(self.payment, request)
        self.assertEqual(type(response), HttpResponse)
        self.assertEqual(self.payment.status, PaymentStatus.CONFIRMED)

    def test_incorrect_custom_token_process_data(self):
        """
        CoinbaseProvider.process_data() checks request custom token
        """
        data = dict(COINBASE_REQUEST)
        data.update({"order": {"custom": "fake"}})
        request = MagicMock()
        request.body = json.dumps(data)
        response = self.provider.process_data(self.payment, request)
        self.assertEqual(type(response), HttpResponseForbidden)

    def test_incorrect_data_process_data(self):
        """
        CoinbaseProvider.process_data() checks request body
        """
        request = MagicMock()
        request.POST = {"id": "1234"}
        response = self.provider.process_data(self.payment, request)
        self.assertEqual(type(response), HttpResponseForbidden)

    @patch("time.time")
    @patch("requests.post")
    def test_provider_returns_checkout_url(self, mocked_post, mocked_time):
        code = "123abc"
        signature = (
            "21d476eff7b2e6cccdfe6deb0c097ba638d5de7e775b303e" "4fdb2f8bfeff72e2"
        )
        url = "https://sandbox.coinbase.com/checkouts/%s" % code
        post = MagicMock()
        post.json = MagicMock(return_value={"button": {"code": code}})
        post.status_code = 200
        mocked_post.return_value = post
        mocked_time.return_value = 1

        form = self.provider.get_form(self.payment)
        self.assertEqual(form.action, url)
        self.assertEqual(
            mocked_post.call_args[1]["headers"]["ACCESS_SIGNATURE"], signature
        )



from uuid import uuid4

from django import forms
from django.template.loader import render_to_string
from django.utils.translation import gettext as _

from .. import PaymentError
from ..forms import CreditCardPaymentFormWithName


class FingerprintWidget(forms.HiddenInput):
    def render(self, name, value, attrs=None, renderer=None):
        final_attrs = dict(attrs or {}, type=self.input_type, name=name)
        final_attrs.update(self.attrs)
        final_attrs["session_id"] = value
        return render_to_string("payments/cybersource_fingerprint.html", final_attrs)


class FingerprintInput(forms.CharField):
    widget = FingerprintWidget
    hidden_widget = FingerprintWidget

    def __init__(self, org_id, merchant_id, fingerprint_url, *args, **kwargs):
        self.org_id = org_id
        self.merchant_id = merchant_id
        self.fingerprint_url = fingerprint_url
        super().__init__(*args, **kwargs)

    def widget_attrs(self, widget):
        attrs = super().widget_attrs(widget=widget)
        attrs["org_id"] = self.org_id
        attrs["merchant_id"] = self.merchant_id
        attrs["fingerprint_url"] = self.fingerprint_url
        return attrs


class PaymentForm(CreditCardPaymentFormWithName):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.provider.org_id:
            try:
                fingerprint_id = self.payment.attrs.fingerprint_session_id
            except KeyError:
                fingerprint_id = str(uuid4())
            self.fields["fingerprint"] = FingerprintInput(
                label=_("fingerprint"),
                org_id=self.provider.org_id,
                initial=fingerprint_id,
                merchant_id=self.provider.merchant_id,
                fingerprint_url=self.provider.fingerprint_url,
            )

    def clean(self):
        cleaned_data = super().clean()
        if not self.errors:
            if self.provider.org_id:
                fingerprint = cleaned_data["fingerprint"]
                self.payment.attrs.fingerprint_session_id = fingerprint
            if not self.payment.transaction_id:
                try:
                    self.provider.charge(self.payment, cleaned_data)
                except PaymentError as e:
                    self._errors["__all__"] = self.error_class([e.args[0]])
            else:
                msg = _("This payment has already been processed.")
                self._errors["__all__"] = self.error_class([msg])
        return cleaned_data



from decimal import Decimal
from unittest import TestCase
from unittest.mock import NonCallableMock
from unittest.mock import patch

from payments import core

from . import PaymentStatus
from .forms import CreditCardPaymentFormWithName
from .forms import PaymentForm
from .models import BasePayment


class TestHelpers(TestCase):
    @patch("payments.core.PAYMENT_HOST", new_callable=NonCallableMock)
    def test_text_get_base_url(self, host):
        host.__str__ = lambda x: "example.com/string"
        self.assertEqual(core.get_base_url(), "https://example.com/string")

    @patch("payments.core.PAYMENT_HOST")
    def test_callable_get_base_url(self, host):
        host.return_value = "example.com/callable"
        self.assertEqual(core.get_base_url(), "https://example.com/callable")


class TestProviderFactory(TestCase):
    def test_provider_factory(self):
        core.provider_factory("default")

    def test_provider_does_not_exist(self):
        self.assertRaises(ValueError, core.provider_factory, "fake_provider")


class Payment(BasePayment):
    """
    Just a concrete model class for testing since instantiating
    an abstract model is deprecated in newer versions of Django.
    """


class TestBasePayment(TestCase):
    def test_payment_attributes(self):
        payment = Payment(extra_data='{"attr1": "test1", "attr2": "test2"}')
        self.assertEqual(payment.attrs.attr1, "test1")
        self.assertEqual(payment.attrs.attr2, "test2")
        self.assertEqual(getattr(payment.attrs, "attr5", None), None)
        self.assertEqual(hasattr(payment.attrs, "attr7"), False)

    def test_capture_with_wrong_status(self):
        payment = Payment(variant="default", status=PaymentStatus.WAITING)
        self.assertRaises(ValueError, payment.capture)

    @patch("payments.dummy.DummyProvider.capture")
    def test_capture_preauth_successfully(self, mocked_capture_method):
        amount = Decimal("20")
        with patch.object(BasePayment, "save") as mocked_save_method:
            mocked_save_method.return_value = None
            mocked_capture_method.return_value = amount

            payment = Payment(variant="default", status=PaymentStatus.PREAUTH)
            payment.capture(amount)

            self.assertEqual(payment.status, PaymentStatus.CONFIRMED)
            self.assertEqual(payment.captured_amount, amount)
        self.assertEqual(mocked_capture_method.call_count, 1)

    @patch("payments.dummy.DummyProvider.capture")
    def test_capture_preauth_without_amount(self, mocked_capture_method):
        amount = None
        with patch.object(BasePayment, "save") as mocked_save_method:
            mocked_save_method.return_value = None
            mocked_capture_method.return_value = amount

            captured_amount = Decimal("100")
            status = PaymentStatus.PREAUTH
            payment = Payment(
                variant="default", status=status, captured_amount=captured_amount
            )
            payment.capture(amount)

            self.assertEqual(payment.status, status)
            self.assertEqual(payment.captured_amount, captured_amount)
        self.assertEqual(mocked_capture_method.call_count, 1)

    def test_release_with_wrong_status(self):
        payment = Payment(variant="default", status=PaymentStatus.WAITING)
        self.assertRaises(ValueError, payment.release)

    @patch("payments.dummy.DummyProvider.release")
    def test_release_preauth_successfully(self, mocked_release_method):
        with patch.object(BasePayment, "save") as mocked_save_method:
            mocked_save_method.return_value = None

            payment = Payment(variant="default", status=PaymentStatus.PREAUTH)
            payment.release()
            self.assertEqual(payment.status, PaymentStatus.REFUNDED)
        self.assertEqual(mocked_release_method.call_count, 1)

    def test_refund_with_wrong_status(self):
        payment = Payment(variant="default", status=PaymentStatus.WAITING)
        self.assertRaises(ValueError, payment.refund)

    def test_refund_too_high_amount(self):
        payment = Payment(
            variant="default",
            status=PaymentStatus.CONFIRMED,
            captured_amount=Decimal("100"),
        )
        self.assertRaises(ValueError, payment.refund, Decimal("200"))

    @patch("payments.dummy.DummyProvider.refund")
    def test_refund_without_amount(self, mocked_refund_method):
        refund_amount = None
        with patch.object(BasePayment, "save") as mocked_save_method:
            mocked_save_method.return_value = None
            mocked_refund_method.return_value = refund_amount

            captured_amount = Decimal("200")
            status = PaymentStatus.CONFIRMED
            payment = Payment(
                variant="default", status=status, captured_amount=captured_amount
            )
            payment.refund(refund_amount)
            self.assertEqual(payment.status, status)
            self.assertEqual(payment.captured_amount, captured_amount)
        self.assertEqual(mocked_refund_method.call_count, 0)

    @patch("payments.dummy.DummyProvider.refund")
    def test_refund_partial_success(self, mocked_refund_method):
        refund_amount = Decimal("100")
        captured_amount = Decimal("200")
        status = PaymentStatus.CONFIRMED
        with patch.object(BasePayment, "save") as mocked_save_method:
            mocked_save_method.return_value = None
            mocked_refund_method.return_value = refund_amount

            payment = Payment(
                variant="default", status=status, captured_amount=captured_amount
            )
            payment.refund(refund_amount)
            self.assertEqual(payment.status, status)
            self.assertEqual(payment.captured_amount, Decimal("100"))
        self.assertEqual(mocked_refund_method.call_count, 1)

    @patch("payments.dummy.DummyProvider.refund")
    def test_refund_fully_success(self, mocked_refund_method):
        refund_amount = Decimal("200")
        captured_amount = Decimal("200")
        with patch.object(BasePayment, "save") as mocked_save_method:
            mocked_save_method.return_value = None
            mocked_refund_method.return_value = refund_amount

            payment = Payment(
                variant="default",
                status=PaymentStatus.CONFIRMED,
                captured_amount=captured_amount,
            )
            payment.refund(refund_amount)
            self.assertEqual(payment.status, PaymentStatus.REFUNDED)
            self.assertEqual(payment.captured_amount, Decimal("0"))
        self.assertEqual(mocked_refund_method.call_count, 1)


class TestCreditCardPaymentForm(TestCase):
    def setUp(self):
        self.data = {
            "name": "John Doe",
            "number": "4716124728800975",
            "expiration_0": "5",
            "expiration_1": "2023",
            "cvv2": "123",
        }

    def test_form_verifies_card_number(self):
        form = CreditCardPaymentFormWithName(data=self.data)
        self.assertTrue(form.is_valid())

    def test_form_raises_error_for_invalid_card_number(self):
        data = dict(self.data)
        data.update({"number": "1112223334445556"})
        form = CreditCardPaymentFormWithName(data=data)
        self.assertFalse(form.is_valid())
        self.assertIn("number", form.errors)

    def test_form_raises_error_for_invalid_cvv2(self):
        data = dict(self.data)
        data.update({"cvv2": "12345"})
        form = CreditCardPaymentFormWithName(data=data)
        self.assertFalse(form.is_valid())
        self.assertIn("cvv2", form.errors)


class TestPaymentForm(TestCase):
    def test_form_contains_hidden_fields(self):
        data = {
            "field1": "value1",
            "field2": "value2",
            "field3": "value3",
            "field4": "value4",
        }

        form = PaymentForm(data=data, hidden_inputs=True)
        self.assertEqual(len(form.fields), len(data))
        self.assertEqual(form.fields["field1"].initial, "value1")


class TestCardIssuer(TestCase):
    def test_mastercard(self):
        self.assertEqual(
            core.get_credit_card_issuer("2720999018275485"),
            ("mastercard", "MasterCard"),
        )
        self.assertEqual(
            core.get_credit_card_issuer("5101395940513451"),
            ("mastercard", "MasterCard"),
        )
        self.assertEqual(
            core.get_credit_card_issuer("5469166706524768"),
            ("mastercard", "MasterCard"),
        )

    def test_visa(self):
        self.assertEqual(
            core.get_credit_card_issuer("4929299255922609"), ("visa", "VISA")
        )
        self.assertEqual(
            core.get_credit_card_issuer("4539883983691685"), ("visa", "VISA")
        )
        self.assertEqual(
            core.get_credit_card_issuer("4916396455393611281"),
            ("visa", "VISA"),
            "19 digit Visa Card",
        )

    def test_discover(self):
        self.assertEqual(
            core.get_credit_card_issuer("6011281400356614"), ("discover", "Discover")
        )
        self.assertEqual(
            core.get_credit_card_issuer("6011223438090674"), ("discover", "Discover")
        )
        self.assertEqual(
            core.get_credit_card_issuer("6011509478386387430"),
            ("discover", "Discover"),
            "19 digit Discover Card",
        )

    def test_amex(self):
        self.assertEqual(
            core.get_credit_card_issuer("341841172626538"), ("amex", "American Express")
        )
        self.assertEqual(
            core.get_credit_card_issuer("348710065929999"), ("amex", "American Express")
        )
        self.assertEqual(
            core.get_credit_card_issuer("341473920579841"), ("amex", "American Express")
        )



import braintree

from .. import PaymentStatus
from ..forms import CreditCardPaymentFormWithName


class BraintreePaymentForm(CreditCardPaymentFormWithName):
    transaction_id = None

    def clean(self):
        data = self.cleaned_data

        if not self.errors and not self.payment.transaction_id:
            result = braintree.Transaction.sale(
                {
                    "amount": str(self.payment.total),
                    "billing": self.get_billing_data(),
                    "credit_card": self.get_credit_card_clean_data(),
                    "customer": self.get_customer_data(),
                    "options": {"submit_for_settlement": False},
                    "order_id": self.payment.description,
                }
            )

            if result.is_success:
                self.transaction_id = result.transaction.id
            else:
                self._errors["__all__"] = self.error_class([result.message])
                self.payment.change_status(PaymentStatus.ERROR)

        return data

    def get_credit_card_clean_data(self):
        if self.cleaned_data:
            return {
                "number": self.cleaned_data.get("number"),
                "cvv": self.cleaned_data.get("cvv2"),
                "cardholder_name": self.cleaned_data.get("name"),
                "expiration_month": self.cleaned_data.get("expiration").month,
                "expiration_year": self.cleaned_data.get("expiration").year,
            }

    def get_billing_data(self):
        return {
            "first_name": self.payment.billing_first_name,
            "last_name": self.payment.billing_last_name,
            "street_address": self.payment.billing_address_1,
            "extended_address": self.payment.billing_address_2,
            "locality": self.payment.billing_city,
            "region": self.payment.billing_country_area,
            "postal_code": self.payment.billing_postcode,
            "country_code_alpha2": self.payment.billing_country_code,
        }

    def get_customer_data(self):
        return {
            "first_name": self.payment.billing_first_name,
            "last_name": self.payment.billing_last_name,
        }

    def save(self):
        braintree.Transaction.submit_for_settlement(self.transaction_id)
        self.payment.transaction_id = self.transaction_id
        self.payment.captured_amount = self.payment.total
        self.payment.change_status(PaymentStatus.CONFIRMED)



from unittest import TestCase
from unittest.mock import MagicMock
from unittest.mock import Mock
from unittest.mock import patch

import pytest

from .. import PaymentStatus
from .. import RedirectNeeded
from . import AuthorizeNetProvider

LOGIN_ID = "abcd1234"
TRANSACTION_KEY = "1234abdd"

PROCESS_DATA = {
    "number": "4007000000027",
    "expiration_0": "5",
    "expiration_1": "2023",
    "cvv2": "123",
}

STATUS_CONFIRMED = "1"
STATUS_DECLINED = "2"
ERROR_PROCESSING = "3"


class Payment(Mock):
    id = 1
    variant = "authorizenet"
    currency = "USD"
    total = 100
    status = PaymentStatus.WAITING
    transaction_id = None
    captured_amount = 0
    message = ""

    def get_process_url(self):
        return "http://example.com"

    def get_failure_url(self):
        return "http://cancel.com"

    def get_success_url(self):
        return "http://success.com"

    def change_status(self, status, message=""):
        self.status = status
        self.message = message


class TestAuthorizeNetProvider(TestCase):
    def setUp(self):
        self.payment = Payment()

    def test_provider_redirects_to_success_on_payment_success(self):
        provider = AuthorizeNetProvider(
            login_id=LOGIN_ID, transaction_key=TRANSACTION_KEY
        )

        response_data = [
            STATUS_CONFIRMED,
            "",
            "",
            "This transaction has been approved.",
            "",
            "",
            "1234",
        ]

        with patch("requests.post") as mocked_post:
            post = MagicMock()
            post.text = "|".join(response_data)
            mocked_post.return_value = post
            with self.assertRaises(RedirectNeeded) as exc:
                provider.get_form(self.payment, data=PROCESS_DATA)
                self.assertEqual(exc.args[0], self.payment.get_success_url())
        self.assertEqual(self.payment.status, PaymentStatus.CONFIRMED)
        self.assertEqual(self.payment.captured_amount, self.payment.total)

    @pytest.mark.skip
    def test_provider_shows_validation_error_message(self):
        provider = AuthorizeNetProvider(
            login_id=LOGIN_ID, transaction_key=TRANSACTION_KEY
        )

        error_msg = "The merchant does not accept this type of credit card."
        response_data = [ERROR_PROCESSING, "", "", error_msg, "", "", "1234"]

        with patch("requests.post") as mocked_post:
            post = MagicMock()
            post.text = "|".join(response_data)
            mocked_post.return_value = post
            form = provider.get_form(self.payment, data=PROCESS_DATA)
            self.assertEqual(form.errors["__all__"][0], error_msg)
            self.assertFalse(form.is_valid())
        self.assertEqual(self.payment.status, "error")
        self.assertEqual(self.payment.captured_amount, 0)
        self.assertEqual(self.payment.message, error_msg)

    @pytest.mark.skip
    def test_provider_shows_rejection_error_message(self):
        provider = AuthorizeNetProvider(
            login_id=LOGIN_ID, transaction_key=TRANSACTION_KEY
        )

        error_msg = " This transaction has been declined."
        response_data = [STATUS_DECLINED, "", "", error_msg, "", "", "1234"]

        with patch("requests.post") as mocked_post:
            post = MagicMock()
            post.text = "|".join(response_data)
            mocked_post.return_value = post
            form = provider.get_form(self.payment, data=PROCESS_DATA)
            self.assertEqual(form.errors["__all__"][0], error_msg)
            self.assertFalse(form.is_valid())
        self.assertEqual(self.payment.status, "rejected")
        self.assertEqual(self.payment.captured_amount, 0)
        self.assertEqual(self.payment.message, error_msg)



from django.http import HttpResponse
from django.shortcuts import get_object_or_404
from django.shortcuts import redirect
from django.template.response import TemplateResponse

from payments import RedirectNeeded
from payments import get_payment_model
from testapp.testmain.forms import TestPaymentForm


def payment_details(request, payment_id):
    """
    Defautl view implemented from docs.
    This view is intented only for interactive testing purposes.
    """
    payment = get_object_or_404(get_payment_model(), id=payment_id)
    try:
        form = payment.get_form(data=request.POST or None)
    except RedirectNeeded as redirect_to:
        return redirect(str(redirect_to))
    return TemplateResponse(request, "payment.html", {"form": form, "payment": payment})


def payment_success(request):
    return HttpResponse("Payment succeeded.")


def payment_failure(request):
    return HttpResponse("Payment failed.")


def create_test_payment(request):
    """
    Creates a basic payment with some default parameters
    to make testing easier.

    If the payment is created successfully, the user is redirected
    to the payment details view where the get_form() method is called,
    or a redirect to the providers link is triggered.
    """
    form = TestPaymentForm(
        initial={"variant": "default", "currency": "USD", "total": 10.0},
        data=request.POST or None,
    )
    if request.method == "POST" and form.is_valid():
        p = form.instance
        p.description = "Product"
        p.save()
        return redirect(f"/test/payment-details/{p.id}")
    return TemplateResponse(request, "create_payment.html", {"form": form})



from django.urls import path

from testapp.testmain import views

urlpatterns = [
    path("payment-details/<int:payment_id>", views.payment_details),
    path("payment-success", views.payment_success),
    path(
        "payment-failure",
        views.payment_failure,
    ),
    path("", views.create_test_payment),
]



from collections import OrderedDict

from django import forms
from django.utils.translation import gettext_lazy as _

from .fields import CreditCardExpiryField
from .fields import CreditCardNameField
from .fields import CreditCardNumberField
from .fields import CreditCardVerificationField


class PaymentForm(forms.Form):
    """
    Payment form, suitable for Django templates.

    When displaying the form remember to use *action* and *method*.
    """

    def __init__(
        self,
        data=None,
        action="",
        method="post",
        provider=None,
        payment=None,
        hidden_inputs=True,
        autosubmit=False,
    ):
        if hidden_inputs and data is not None:
            super().__init__(auto_id=False)
            for key, val in data.items():
                widget = forms.widgets.HiddenInput()
                self.fields[key] = forms.CharField(initial=val, widget=widget)
        else:
            super().__init__(data=data)
        self.action = action
        self.autosubmit = autosubmit
        self.method = method
        self.provider = provider
        self.payment = payment


class CreditCardPaymentForm(PaymentForm):
    number = CreditCardNumberField(label=_("Card Number"), max_length=32, required=True)
    expiration = CreditCardExpiryField()
    cvv2 = CreditCardVerificationField(
        label=_("CVV2 Security Number"),
        required=False,
        help_text=_(
            "Last three digits located on the back of your card."
            " For American Express the four digits found on the front side."
        ),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(hidden_inputs=False, *args, **kwargs)
        if hasattr(self, "VALID_TYPES"):
            self.fields["number"].valid_types = self.VALID_TYPES


class CreditCardPaymentFormWithName(CreditCardPaymentForm):
    name = CreditCardNameField(label=_("Name on Credit Card"), max_length=128)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        name_field = self.fields.pop("name")
        fields = OrderedDict({"name": name_field})
        fields.update(self.fields)
        self.fields = fields
